from typing import List, Optional, Dict, Tuple

from typing_extensions import Literal

from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.group import AwsResourceGroup
from phidata.infra.aws.resource.types import (
    AwsResourceInstallOrder,
    AwsResourceType,
)
from phidata.utils.log import logger


def get_install_weight_for_aws_resource(
    aws_resource_type: AwsResource, resource_group_weight: int = 100
) -> int:
    """Function which returns the install weight for an AwsResource"""

    resource_type_class_name = aws_resource_type.__class__.__name__
    if resource_type_class_name in AwsResourceInstallOrder.keys():
        install_weight = AwsResourceInstallOrder[resource_type_class_name]
        final_weight = resource_group_weight * install_weight
        # logger.debug(
        #     "Resource type: {} | Install weight: {}".format(
        #         resource_type_class_name,
        #         install_weight,
        #     )
        # )
        return final_weight

    return 5000


def get_aws_resources_from_group(
    aws_resource_group: AwsResourceGroup,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
) -> List[AwsResource]:
    """Parses the AwsResourceGroup and returns an array of AwsResources
    after applying the name & type filters. This function also flattens any
    List[AwsResource] attributes. Eg: it will flatten s3_buckets

    Args:
        aws_resource_group:
        name_filter: filter AwsResources by name
        type_filter: filter AwsResources by type

    Returns:
        List[AwsResource]: List of filtered and flattened Aws Resources
    """

    # List of resources to return
    aws_resources: List[AwsResource] = []
    # logger.debug(f"Flattening {aws_resource_group.name}")

    # populate the aws_resources list with the resources
    # Loop over each key of the AwsResourceGroup model
    for resource, resource_data in aws_resource_group.__dict__.items():
        # logger.debug("resource: {}".format(resource))
        # logger.debug("resource_data: {}".format(resource_data))

        # Check if the resource is a single AwsResource or a List[AwsResource]
        # If it is a List[AwsResource], flatten the resources, verify each element
        # of the list is a subclass of AwsResource and add to the aws_resources list
        if isinstance(resource_data, list):
            for _r in resource_data:
                if isinstance(_r, AwsResource):
                    rn = _r.get_resource_name()
                    rt = _r.get_resource_type()
                    if name_filter is not None and rn is not None:
                        # logger.debug(f"name_filter: {name_filter.lower()}")
                        # logger.debug(f"resource name: {rn.lower()}")
                        if name_filter.lower() not in rn.lower():
                            logger.debug(f"  -*- skipping {rt}:{rn}")
                            continue

                    if type_filter is not None and rt is not None:
                        # logger.debug(f"type_filter: {type_filter.lower()}")
                        # logger.debug(f"class: {rt.lower()}")
                        if type_filter.lower() != rt.lower():
                            logger.debug(f"  -*- skipping {rt}:{rn}")
                            continue
                    aws_resources.append(_r)  # type: ignore

        # If its a single resource, verify that the resource is a subclass of
        # AwsResource and add it to the aws_resources list
        elif isinstance(resource_data, AwsResource):
            rn = resource_data.get_resource_name()
            rt = resource_data.get_resource_type()
            if name_filter is not None and rn is not None:
                # logger.debug(f"name_filter: {name_filter.lower()}")
                # logger.debug(f"resource name: {rn.lower()}")
                if name_filter.lower() not in rn.lower():
                    # logger.debug(f"skipping {rt}:{rn}")
                    continue

            if type_filter is not None and rt is not None:
                # logger.debug(f"type_filter: {type_filter.lower()}")
                # logger.debug(f"class: {rt.lower()}")
                if type_filter.lower() != rt.lower():
                    # logger.debug(f"skipping {rt}:{rn}")
                    continue
            aws_resources.append(resource_data)  # type: ignore

    return aws_resources


# def dedup_resource_types(
#     aws_resources: Optional[List[AwsResource]] = None,
# ) -> Optional[Set[Type[AwsResource]]]:
#     """Takes a list of AwsResources and returns a Set of AwsResource classes.
#     Each AwsResource classes is represented by the Type[resources.AwsResource] type.
#     From python docs:
#         A variable annotated with Type[AwsResource] may accept values that are classes.
#         Acceptable classes are the AwsResource class + subclasses.
#     """
#     if aws_resources:
#         active_resource_types: Set[Type[AwsResource]] = set()
#         for resource in aws_resources:
#             active_resource_types.add(resource.__class__)
#             # logger.debug(f"Gathering: {resource.get_resource_name()}")
#             # logger.debug(f"Resource Type: {resource_type}")
#         logger.debug("Active Resource Types: {}".format(active_resource_types))
#         return active_resource_types
#     return None


def dedup_aws_resources(
    aws_resources_with_weight: List[Tuple[AwsResource, int]],
) -> List[Tuple[AwsResource, int]]:
    if aws_resources_with_weight is None:
        raise ValueError

    deduped_resources: List[Tuple[AwsResource, int]] = []
    prev_rsrc: Optional[AwsResource] = None
    prev_weight: Optional[int] = None
    for rsrc, weight in aws_resources_with_weight:
        # First item of loop
        if prev_rsrc is None:
            prev_rsrc = rsrc
            prev_weight = weight
            deduped_resources.append((rsrc, weight))
            continue

        # Compare resources with same weight only
        if weight == prev_weight:
            if (
                rsrc.get_resource_type() == prev_rsrc.get_resource_type()
                and rsrc.get_resource_name() == prev_rsrc.get_resource_name()
                and rsrc.get_resource_name() is not None
            ):
                # If resource type and name are the same, skip the resource
                # Note: resource.name cannot be None
                continue

        # If the loop hasn't been continued by the if blocks above,
        # add the resource and weight to the deduped_resources list
        deduped_resources.append((rsrc, weight))
        # update the previous resource for comparison
        prev_rsrc = rsrc
        prev_weight = weight
    return deduped_resources


def filter_and_flatten_aws_resource_groups(
    aws_resource_groups: Dict[str, AwsResourceGroup],
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    app_filter: Optional[str] = None,
    sort_order: Literal["create", "delete"] = "create",
) -> Optional[List[AwsResource]]:
    """This function flattens the aws_resource_group and returns a filtered list of
    AwsResources sorted in the order requested. create == install order, delete == reverse

    Args:
        aws_resource_groups: AwsResourceGroup
        name_filter: Filter AwsResources by name
        type_filter: Filter AwsResources by type
        app_filter: Filter AwsResources by aws resource group name
        sort_order: create or delete

    Returns:
        List[AwsResource]: List of filtered Aws Resources
    """

    logger.debug("Filtering & Flattening AwsResourceGroups")

    # Step 1: Create aws_resource_list_with_weight
    # A List of Tuples where each tuple is a (AwsResource, Resource Group Weight)
    # The reason for creating this list is so that we can sort the AwsResources
    # based on their resource group weight using get_install_weight_for_aws_resource
    aws_resource_list_with_weight: List[Tuple[AwsResource, int]] = []
    if aws_resource_groups:
        # Iterate through aws_resource_groups
        for aws_rg_name, aws_rg in aws_resource_groups.items():
            # logger.debug("aws_rg_name: {}".format(aws_rg_name))
            # logger.debug("aws_rg: {}".format(aws_rg))

            # skip disabled AwsResourceGroups
            if not aws_rg.enabled:
                continue

            # skip groups not matching app_filter if provided
            if app_filter is not None and aws_rg_name is not None:
                # logger.debug(f"matching app_filter: {app_filter}")
                # logger.debug(f"with aws_rg_name: {aws_rg_name}")
                if app_filter.lower() not in aws_rg_name.lower():
                    logger.debug(f"  -*- skipping {aws_rg_name}")
                    continue

            aws_resources = get_aws_resources_from_group(
                aws_rg, name_filter, type_filter
            )
            if aws_resources:
                for _aws_rsrc in aws_resources:
                    aws_resource_list_with_weight.append((_aws_rsrc, aws_rg.weight))

    # Sort the resources in install order
    if sort_order == "create":
        aws_resource_list_with_weight.sort(
            key=lambda x: get_install_weight_for_aws_resource(x[0], x[1])
        )
    elif sort_order == "delete":
        aws_resource_list_with_weight.sort(
            key=lambda x: get_install_weight_for_aws_resource(x[0], x[1]), reverse=True
        )

    # De-duplicate AwsResources
    deduped_aws_resources: List[Tuple[AwsResource, int]] = dedup_aws_resources(
        aws_resource_list_with_weight
    )
    # deduped_aws_resources = aws_resource_list_with_weight
    # logger.debug("AwsResource list")
    # logger.debug(
    #     "\n".join(
    #         "Name: {}, Type: {}, Weight: {}".format(
    #             rsrc.name, rsrc.resource_type, weight
    #         )
    #         for rsrc, weight in deduped_aws_resources
    #     )
    # )

    # drop the weight from the deduped_aws_resources tuple
    filtered_aws_resources: List[AwsResource] = [x[0] for x in deduped_aws_resources]
    # logger.debug("filtered_aws_resources: {}".format(filtered_aws_resources))
    return filtered_aws_resources



from typing import List, Optional

from pydantic import BaseModel

from phidata.infra.aws.resource.acm.certificate import AcmCertificate
from phidata.infra.aws.resource.cloudformation.stack import CloudFormationStack
from phidata.infra.aws.resource.ec2.volume import EbsVolume
from phidata.infra.aws.resource.eks.cluster import EksCluster
from phidata.infra.aws.resource.eks.fargate_profile import EksFargateProfile
from phidata.infra.aws.resource.eks.node_group import EksNodeGroup
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.infra.aws.resource.iam.policy import IamPolicy
from phidata.infra.aws.resource.glue.crawler import GlueCrawler
from phidata.infra.aws.resource.s3.bucket import S3Bucket
from phidata.infra.aws.resource.emr.cluster import EmrCluster
from phidata.infra.aws.resource.rds.db_cluster import DbCluster
from phidata.infra.aws.resource.rds.db_subnet_group import DbSubnetGroup


class AwsResourceGroup(BaseModel):
    """The AwsResourceGroup class contains the data for all AwsResources"""

    name: str = "default"
    enabled: bool = True
    weight: int = 100

    iam_roles: Optional[List[IamRole]] = None
    iam_policies: Optional[List[IamPolicy]] = None
    acm_certificates: Optional[List[AcmCertificate]] = None
    s3_buckets: Optional[List[S3Bucket]] = None
    db_subnet_groups: Optional[List[DbSubnetGroup]] = None
    dbs: Optional[List[DbCluster]] = None
    volumes: Optional[List[EbsVolume]] = None
    cloudformation_stacks: Optional[List[CloudFormationStack]] = None
    eks_cluster: Optional[EksCluster] = None
    eks_fargate_profiles: Optional[List[EksFargateProfile]] = None
    eks_nodegroups: Optional[List[EksNodeGroup]] = None
    crawlers: Optional[List[GlueCrawler]] = None
    emr: Optional[List[EmrCluster]] = None



from typing import List, Optional, Dict
from typing_extensions import Literal

from kubernetes.client import CoreV1Api
from kubernetes.client.models.v1_persistent_volume import V1PersistentVolume
from kubernetes.client.models.v1_persistent_volume_list import V1PersistentVolumeList
from kubernetes.client.models.v1_persistent_volume_spec import V1PersistentVolumeSpec
from kubernetes.client.models.v1_status import V1Status
from pydantic import Field

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.enums.pv import PVAccessMode
from phidata.infra.k8s.resource.base import K8sResource, K8sObject
from phidata.infra.k8s.resource.core.v1.volume_source import (
    GcePersistentDiskVolumeSource,
    LocalVolumeSource,
    HostPathVolumeSource,
)
from phidata.infra.k8s.resource.core.v1.volume_node_affinity import VolumeNodeAffinity
from phidata.utils.log import logger


class PersistentVolumeSpec(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#persistentvolumeclaim-v1-core
    """

    resource_type: str = "PersistentVolumeSpec"

    # AccessModes contains all ways the volume can be mounted.
    # More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes#access-modes
    access_modes: List[PVAccessMode] = Field(..., alias="accessModes")
    # A description of the persistent volume's resources and capacity.
    # More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes#capacity
    capacity: Optional[Dict[str, str]] = None
    # A list of mount options, e.g. ["ro", "soft"]. Not validated - mount will simply fail if one is invalid.
    # More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes/#mount-options
    mount_options: Optional[List[str]] = Field(None, alias="mountOptions")
    # NodeAffinity defines constraints that limit what nodes this volume can be accessed from.
    # This field influences the scheduling of pods that use this volume.
    node_affinity: Optional[VolumeNodeAffinity] = Field(None, alias="nodeAffinity")
    # What happens to a persistent volume when released from its claim.
    # Valid options are Retain (default for manually created PersistentVolumes)
    #   Delete (default for dynamically provisioned PersistentVolumes)
    #   Recycle (deprecated). Recycle must be supported by the volume plugin underlying this PersistentVolume.
    # More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes#reclaiming
    # Possible enum values:
    #   - `"Delete"` means the volume will be deleted from Kubernetes on release from its claim.
    #   - `"Recycle"` means the volume will be recycled back into the pool of unbound persistent volumes
    #           on release from its claim.
    #   - `"Retain"` means the volume will be left in its current phase (Released) for manual reclamation
    #           by the administrator.
    #   The default policy is Retain.
    persistent_volume_reclaim_policy: Optional[
        Literal["Delete", "Recycle", "Retain"]
    ] = Field(None, alias="persistentVolumeReclaimPolicy")
    # Name of StorageClass to which this persistent volume belongs.
    # Empty value means that this volume does not belong to any StorageClass.
    storage_class_name: Optional[str] = Field(None, alias="storageClassName")
    # volumeMode defines if a volume is intended to be used with a formatted filesystem or to remain in raw block state.
    # Value of Filesystem is implied when not included in spec.
    volume_mode: Optional[str] = Field(None, alias="volumeMode")

    ## Volume Sources
    # Local represents directly-attached storage with node affinity
    local: Optional[LocalVolumeSource] = None
    # HostPath represents a directory on the host. Provisioned by a developer or tester.
    # This is useful for single-node development and testing only!
    # On-host storage is not supported in any way and WILL NOT WORK in a multi-node cluster.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#hostpath
    host_path: Optional[HostPathVolumeSource] = Field(None, alias="hostPath")
    # GCEPersistentDisk represents a GCE Disk resource that is attached to a
    # kubelet's host machine and then exposed to the pod. Provisioned by an admin.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#gcepersistentdisk
    gce_persistent_disk: Optional[GcePersistentDiskVolumeSource] = Field(
        None, alias="gcePersistentDisk"
    )

    def get_k8s_object(
        self,
    ) -> V1PersistentVolumeSpec:

        # Return a V1PersistentVolumeSpec object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_persistent_volume_spec.py
        _v1_persistent_volume_spec = V1PersistentVolumeSpec(
            access_modes=self.access_modes,
            capacity=self.capacity,
            mount_options=self.mount_options,
            persistent_volume_reclaim_policy=self.persistent_volume_reclaim_policy,
            storage_class_name=self.storage_class_name,
            volume_mode=self.volume_mode,
            local=self.local.get_k8s_object() if self.local else None,
            host_path=self.host_path.get_k8s_object() if self.host_path else None,
            gce_persistent_disk=self.gce_persistent_disk.get_k8s_object()
            if self.gce_persistent_disk
            else None,
            node_affinity=self.node_affinity.get_k8s_object()
            if self.node_affinity
            else None,
        )
        return _v1_persistent_volume_spec


class PersistentVolume(K8sResource):
    """
    A PersistentVolume (PV) is a piece of storage in the cluster that has been provisioned by an administrator
    or dynamically provisioned using Storage Classes.

    In Kubernetes, each container can read and write to its own, isolated filesystem.
    But, data on that filesystem will be destroyed when the container is restarted.
    To solve this, Kubernetes has volumes. Volumes let your pod write to a filesystem that exists as long as the pod exists.
    Volumes also let you share data between containers in the same pod.
    But, data in that volume will be destroyed when the pod is restarted.
    To solve this, Kubernetes has persistent volumes. Persistent volumes are long-term storage in your Kubernetes cluster.
    Persistent volumes exist beyond containers, pods, and nodes.

    A pod uses a persistent volume claim to to get read and write access to the persistent volume.

    References:
        * Docs:
            https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#persistentvolume-v1-core
            https://kubernetes.io/docs/concepts/storage/persistent-volumes/
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_persistent_volume.py
    """

    resource_type: str = "PersistentVolume"

    spec: PersistentVolumeSpec

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = ["spec"]

    def get_k8s_object(self) -> V1PersistentVolume:
        """Creates a body for this PVC"""

        # Return a V1PersistentVolume object to create a ClusterRole
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_persistent_volume.py
        _v1_persistent_volume = V1PersistentVolume(
            api_version=self.api_version,
            kind=self.kind,
            metadata=self.metadata.get_k8s_object(),
            spec=self.spec.get_k8s_object(),
        )
        return _v1_persistent_volume

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1PersistentVolume]]:
        """Reads PVCs from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: NOT USED.
        """
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        pv_list: Optional[V1PersistentVolumeList] = core_v1_api.list_persistent_volume(
            **kwargs
        )
        pvs: Optional[List[V1PersistentVolume]] = None
        if pv_list:
            pvs = pv_list.items
            logger.debug(f"pvs: {pvs}")
            logger.debug(f"pvs type: {type(pvs)}")
        return pvs

    def _create(self, k8s_client: K8sApiClient) -> bool:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        k8s_object: V1PersistentVolume = self.get_k8s_object()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_persistent_volume: V1PersistentVolume = core_v1_api.create_persistent_volume(
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Created: {}".format(v1_persistent_volume))
        if v1_persistent_volume.metadata.creation_timestamp is not None:
            logger.debug("PV Created")
            self.active_resource = v1_persistent_volume
            self.active_resource_class = V1PersistentVolume
            return True
        logger.error("PV could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1PersistentVolume]:
        """Returns the "Active" PVC from the cluster"""

        active_resource: Optional[V1PersistentVolume] = None
        active_resources: Optional[List[V1PersistentVolume]] = self.get_from_cluster(
            k8s_client=k8s_client,
        )
        # logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {_pv.metadata.name: _pv for _pv in active_resources}

        pv_name = self.get_resource_name()
        if pv_name in active_resources_dict:
            active_resource = active_resources_dict[pv_name]
            self.active_resource = active_resource
            self.active_resource_class = V1PersistentVolume
            logger.debug(f"Found active {pv_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        pv_name = self.get_resource_name()
        k8s_object: V1PersistentVolume = self.get_k8s_object()

        logger.debug("Updating: {}".format(pv_name))
        v1_persistent_volume: V1PersistentVolume = core_v1_api.patch_persistent_volume(
            name=pv_name,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Updated:\n{}".format(pformat(v1_persistent_volume.to_dict(), indent=2)))
        if v1_persistent_volume.metadata.creation_timestamp is not None:
            logger.debug("PV Updated")
            self.active_resource = v1_persistent_volume
            self.active_resource_class = V1PersistentVolume
            return True
        logger.error("PV could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        pv_name = self.get_resource_name()

        logger.debug("Deleting: {}".format(pv_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1Status = core_v1_api.delete_persistent_volume(
            name=pv_name,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("delete_status: {}".format(pformat(delete_status, indent=2)))
        if delete_status.status == "Success":
            logger.debug("PV Deleted")
            return True
        logger.error("PV could not be deleted")
        return False



from typing import Optional, Any

from phidata.asset.table.sql import SqlTable
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger
from phidata.task import PythonTask, PythonTaskArgs


class PlotSqlQueryArgs(PythonTaskArgs):
    query: str
    sql_table: SqlTable
    show_sample_data: bool = False
    create_db_engine_from_conn_id: bool = True


class PlotSqlQuery(PythonTask):
    def __init__(
        self,
        query: str,
        sql_table: SqlTable,
        show_sample_data: bool = False,
        name: str = "plot_sql_query",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: PlotSqlQueryArgs = PlotSqlQueryArgs(
                query=query,
                sql_table=sql_table,
                show_sample_data=show_sample_data,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=run_sql_query,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def query(self) -> Optional[str]:
        return self.args.query

    @query.setter
    def query(self, query: str) -> None:
        if query is not None:
            self.args.query = query

    @property
    def sql_table(self) -> Optional[SqlTable]:
        return self.args.sql_table

    @sql_table.setter
    def sql_table(self, sql_table: SqlTable) -> None:
        if sql_table is not None:
            self.args.sql_table = sql_table


def run_sql_query_pandas(args: PlotSqlQueryArgs) -> bool:

    import pandas as pd
    import matplotlib.pyplot as plt

    query: str = args.query
    sql_table: SqlTable = args.sql_table

    print_info("Running Query:\n{}".format(query))
    df: Optional[pd.DataFrame] = sql_table.run_sql_query(query)

    if df is None:
        print_error("Could not run query")
        return False

    if args.show_sample_data:
        print_info("Sample data:\n{}".format(df.head()))
        print_info("Plotting")
        df.plot("ds", "active_users")
        plt.show(block=False)
        plt.pause(3)
        plt.close()
    return True


def run_sql_query(**kwargs) -> bool:
    args: PlotSqlQueryArgs = PlotSqlQueryArgs(**kwargs)
    return run_sql_query_pandas(args)



from collections import OrderedDict
from typing import Any, Dict, List, Optional, Type, Set, cast, Tuple

from phidata.infra.aws.args import AwsArgs
from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.exceptions import (
    AwsArgsException,
    AwsApiClientException,
    AwsResourceCreationFailedException,
)
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.group import AwsResourceGroup
from phidata.infra.aws.resource.types import AwsResourceType
from phidata.infra.aws.resource.utils import (
    get_aws_resources_from_group,
    filter_and_flatten_aws_resource_groups,
)
from phidata.utils.cli_console import (
    print_info,
    print_info,
    print_heading,
    print_subheading,
    print_error,
    confirm_yes_no,
)
from phidata.utils.common import is_empty
from phidata.utils.log import logger


class AwsWorker:
    """This class interacts with the Aws API."""

    def __init__(self, aws_args: AwsArgs) -> None:
        self.aws_args: AwsArgs = aws_args
        self.aws_api_client: AwsApiClient = AwsApiClient(
            aws_region=self.aws_args.aws_region,
            aws_profile=self.aws_args.aws_profile,
        )
        self.aws_resources: Optional[Dict[str, AwsResourceGroup]] = None
        logger.debug(f"**-+-** AwsWorker created")

    def is_client_initialized(self) -> bool:
        return self.aws_api_client.is_initialized()

    def are_resources_initialized(self) -> bool:
        if self.aws_resources is not None and len(self.aws_resources) > 0:
            return True
        return False

    def are_resources_active(self) -> bool:
        # TODO: fix this
        return False

    ######################################################
    ## Init Resources
    ######################################################

    def init_resources(self) -> None:
        """
        This function populates the self.aws_resources dictionary.
        """
        logger.debug("-*- Initializing AwsResourceGroup")

        if self.aws_resources is None:
            self.aws_resources = OrderedDict()

        rgs_to_build = self.aws_args.resources
        if rgs_to_build is not None and isinstance(rgs_to_build, list):
            for resource_group in rgs_to_build:
                self.aws_resources[resource_group.name] = resource_group

        logger.debug(f"# AwsResources built: {len(self.aws_resources)}")

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Creating AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return False

        # A list of tuples with 3 parts
        #   1. Resource group name
        #   2. Resource group weight
        #   3. List of Resources in group after filters
        aws_resources_to_create: List[Tuple[str, int, List[AwsResource]]] = []
        for aws_rg_name, aws_rg in self.aws_resources.items():
            logger.debug(f"Processing: {aws_rg_name}")

            # skip disabled AwsResourceTypeGroups
            if not aws_rg.enabled:
                continue

            # skip groups not matching app_filter if provided
            if app_filter is not None:
                if app_filter.lower() not in aws_rg_name.lower():
                    logger.debug(f"Skipping {aws_rg_name}")
                    continue

            aws_resources_in_group = get_aws_resources_from_group(
                aws_resource_group=aws_rg,
                name_filter=name_filter,
                type_filter=type_filter,
            )
            num_resources_in_group = len(aws_resources_in_group)
            logger.debug(f"Matched {num_resources_in_group} resources")
            if num_resources_in_group > 0:
                aws_resources_to_create.append(
                    (aws_rg_name, aws_rg.weight, aws_resources_in_group)
                )

        if len(aws_resources_to_create) == 0:
            print_subheading("No AwsResource to create")
            return True

        aws_resources_to_create_sorted: List[
            Tuple[str, int, List[AwsResource]]
        ] = sorted(
            aws_resources_to_create,
            key=lambda x: x[1],
        )

        # Validate resources to be created
        group_number = 1
        resource_number = 0
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for (
                group_name,
                group_weight,
                resource_list,
            ) in aws_resources_to_create_sorted:
                print_subheading(f"\n{group_number}. {group_name}")
                resource_number += len(resource_list)
                group_number += 1
                for resource in resource_list:
                    print_info(
                        f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                    )
            if self.aws_args.aws_region:
                print_info(f"\nRegion: {self.aws_args.aws_region}")
            if self.aws_args.aws_profile:
                print_info(f"Profile: {self.aws_args.aws_profile}")
            print_info(f"\nTotal {resource_number} resources")
            confirm = confirm_yes_no("\nConfirm deploy")
            if not confirm:
                print_info("Skipping deploy")
                return False

        # track the total number of AwsResource to create for validation
        num_resources_to_create: int = 0
        num_resources_created: int = 0

        for (group_name, group_weight, resource_list) in aws_resources_to_create_sorted:
            print_subheading(f"\n-*- {group_name}")
            num_resources_to_create += len(resource_list)
            for resource in resource_list:
                if resource:
                    print_info(
                        f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
                    )
                    # logger.debug(resource)
                    try:
                        _resource_created = resource.create(
                            aws_client=self.aws_api_client
                        )
                        if _resource_created:
                            num_resources_created += 1
                    except Exception as e:
                        logger.error(
                            f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be created."
                        )
                        logger.error("Error: {}".format(e))
                        logger.error(
                            "Skipping resource creation, please fix and try again..."
                        )

        print_info(
            f"\n# Resources created: {num_resources_created}/{num_resources_to_create}"
        )
        if num_resources_to_create == num_resources_created:
            return True

        logger.error(
            f"Resources created: {num_resources_created} do not match Resources required: {num_resources_to_create}"
        )
        return False

    def create_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Creating AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return

        # A list of tuples with 3 parts
        #   1. Resource group name
        #   2. Resource group weight
        #   3. List of Resources in group after filters
        aws_resources_to_create: List[Tuple[str, int, List[AwsResource]]] = []
        for aws_rg_name, aws_rg in self.aws_resources.items():
            logger.debug(f"Processing: {aws_rg_name}")

            # skip disabled AwsResourceGroups
            if not aws_rg.enabled:
                continue

            # skip groups not matching app_filter if provided
            if app_filter is not None:
                if app_filter.lower() not in aws_rg_name.lower():
                    logger.debug(f"Skipping {aws_rg_name}")
                    continue

            aws_resources_in_group = get_aws_resources_from_group(
                aws_resource_group=aws_rg,
                name_filter=name_filter,
                type_filter=type_filter,
            )
            num_resources_in_group = len(aws_resources_in_group)
            logger.debug(f"Matched {num_resources_in_group} resources")
            if num_resources_in_group > 0:
                aws_resources_to_create.append(
                    (aws_rg_name, aws_rg.weight, aws_resources_in_group)
                )

        if len(aws_resources_to_create) == 0:
            print_info("No AwsResources to create")
            return

        aws_resources_to_create_sorted: List[
            Tuple[str, int, List[AwsResource]]
        ] = sorted(
            aws_resources_to_create,
            key=lambda x: x[1],
        )

        group_number = 1
        num_resources_to_create: int = 0
        print_heading("--**-- Aws resources:")
        for (group_name, group_weight, resource_list) in aws_resources_to_create_sorted:
            print_subheading(f"\n{group_number}. {group_name}")
            num_resources_to_create += len(resource_list)
            group_number += 1
            for resource in resource_list:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
        if self.aws_args.aws_region:
            print_info(f"\nRegion: {self.aws_args.aws_region}")
        if self.aws_args.aws_profile:
            print_info(f"Profile: {self.aws_args.aws_profile}")
        print_info(f"\nTotal {num_resources_to_create} resources")

    ######################################################
    ## Read Resources
    ######################################################

    def get_resource_groups(self) -> Optional[Dict[str, AwsResourceGroup]]:

        logger.debug("-*- Getting AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No Resources available")
                return {}

        return self.aws_resources

    def read_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> Optional[List[AwsResource]]:

        logger.debug("-*- Getting AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return None

        aws_resources: Optional[
            List[AwsResource]
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=self.aws_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )
        return aws_resources

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Deleting AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return False

        # A list of tuples with 3 parts
        #   1. Resource group name
        #   2. Resource group weight
        #   3. List of Resources in group after filters
        aws_resources_to_delete: List[Tuple[str, int, List[AwsResource]]] = []
        for aws_rg_name, aws_rg in self.aws_resources.items():
            logger.debug(f"Processing: {aws_rg_name}")

            # skip disabled AwsResourceGroups
            if not aws_rg.enabled:
                continue

            # skip groups not matching app_filter if provided
            if app_filter is not None:
                if app_filter.lower() not in aws_rg_name.lower():
                    logger.debug(f"Skipping {aws_rg_name}")
                    continue

            aws_resources_in_group = get_aws_resources_from_group(
                aws_resource_group=aws_rg,
                name_filter=name_filter,
                type_filter=type_filter,
            )
            num_resources_in_group = len(aws_resources_in_group)
            logger.debug(f"Matched {num_resources_in_group} resources")
            if num_resources_in_group > 0:
                aws_resources_to_delete.append(
                    (aws_rg_name, aws_rg.weight, aws_resources_in_group)
                )

        if len(aws_resources_to_delete) == 0:
            print_subheading("No AwsResources to delete")
            return True

        aws_resources_to_delete_sorted: List[
            Tuple[str, int, List[AwsResource]]
        ] = sorted(
            aws_resources_to_delete,
            key=lambda x: x[1],
            reverse=True,
        )

        # Validate resources to be created
        group_number = 1
        resource_number = 0
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for (
                group_name,
                group_weight,
                resource_list,
            ) in aws_resources_to_delete_sorted:
                print_subheading(f"\n{group_number}. {group_name}")
                resource_number += len(resource_list)
                group_number += 1
                for resource in resource_list:
                    print_info(
                        f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                    )
            if self.aws_args.aws_region:
                print_info(f"\nRegion: {self.aws_args.aws_region}")
            if self.aws_args.aws_profile:
                print_info(f"Profile: {self.aws_args.aws_profile}")
            print_info(f"\nTotal {resource_number} resources")
            confirm = confirm_yes_no("\nConfirm delete")
            if not confirm:
                print_info("Skipping delete")
                return False

        # track the total number of AwsResources to create for validation
        num_resources_to_delete: int = 0
        num_resources_deleted: int = 0

        for (group_name, group_weight, resource_list) in aws_resources_to_delete_sorted:
            print_subheading(f"\n-*- {group_name}")
            num_resources_to_delete += len(resource_list)
            for resource in resource_list:
                if resource:
                    print_info(
                        f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
                    )
                    # logger.debug(resource)
                    try:
                        _resource_created = resource.delete(
                            aws_client=self.aws_api_client
                        )
                        if _resource_created:
                            num_resources_deleted += 1
                    except Exception as e:
                        logger.error(
                            f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be created."
                        )
                        logger.error("Error: {}".format(e))
                        logger.error(
                            "Skipping resource creation, please fix and try again..."
                        )

        print_info(
            f"\n# Resources deleted: {num_resources_deleted}/{num_resources_to_delete}"
        )
        if num_resources_to_delete == num_resources_deleted:
            return True

        logger.error(
            f"Resources deleted: {num_resources_deleted} do not match resources required: {num_resources_to_delete}"
        )
        return False

    def delete_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Deleting AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return

        # A list of tuples with 3 parts
        #   1. Resource group name
        #   2. Resource group weight
        #   3. List of Resources in group after filters
        aws_resources_to_delete: List[Tuple[str, int, List[AwsResource]]] = []
        for aws_rg_name, aws_rg in self.aws_resources.items():
            logger.debug(f"Processing: {aws_rg_name}")

            # skip disabled AwsResourceGroups
            if not aws_rg.enabled:
                continue

            # skip groups not matching app_filter if provided
            if app_filter is not None:
                if app_filter.lower() not in aws_rg_name.lower():
                    logger.debug(f"Skipping {aws_rg_name}")
                    continue

            aws_resources_in_group = get_aws_resources_from_group(
                aws_resource_group=aws_rg,
                name_filter=name_filter,
                type_filter=type_filter,
            )
            num_resources_in_group = len(aws_resources_in_group)
            logger.debug(f"Matched {num_resources_in_group} resources")
            if num_resources_in_group > 0:
                aws_resources_to_delete.append(
                    (aws_rg_name, aws_rg.weight, aws_resources_in_group)
                )

        if len(aws_resources_to_delete) == 0:
            print_info("No AwsResources to create")
            return

        aws_resources_to_delete_sorted: List[
            Tuple[str, int, List[AwsResource]]
        ] = sorted(
            aws_resources_to_delete,
            key=lambda x: x[1],
            reverse=True,
        )

        group_number = 1
        num_resources_to_delete: int = 0
        print_heading("--**-- Aws resources:")
        for (group_name, group_weight, resource_list) in aws_resources_to_delete_sorted:
            print_subheading(f"\n{group_number}. {group_name}")
            num_resources_to_delete += len(resource_list)
            group_number += 1
            for resource in resource_list:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
        if self.aws_args.aws_region:
            print_info(f"\nRegion: {self.aws_args.aws_region}")
        if self.aws_args.aws_profile:
            print_info(f"Profile: {self.aws_args.aws_profile}")
        print_info(f"\nTotal {num_resources_to_delete} resources")

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Patching AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return False

        # A list of tuples with 3 parts
        #   1. Resource group name
        #   2. Resource group weight
        #   3. List of Resources in group after filters
        aws_resources_to_patch: List[Tuple[str, int, List[AwsResource]]] = []
        for aws_rg_name, aws_rg in self.aws_resources.items():
            logger.debug(f"Processing: {aws_rg_name}")

            # skip disabled AwsResourceGroups
            if not aws_rg.enabled:
                continue

            # skip groups not matching app_filter if provided
            if app_filter is not None:
                if app_filter.lower() not in aws_rg_name.lower():
                    logger.debug(f"Skipping {aws_rg_name}")
                    continue

            aws_resources_in_group = get_aws_resources_from_group(
                aws_resource_group=aws_rg,
                name_filter=name_filter,
                type_filter=type_filter,
            )
            num_resources_in_group = len(aws_resources_in_group)
            logger.debug(f"Matched {num_resources_in_group} resources")
            if num_resources_in_group > 0:
                aws_resources_to_patch.append(
                    (aws_rg_name, aws_rg.weight, aws_resources_in_group)
                )

        if len(aws_resources_to_patch) == 0:
            print_subheading("No AwsResources to patch")
            return True

        aws_resources_to_patch_sorted: List[
            Tuple[str, int, List[AwsResource]]
        ] = sorted(
            aws_resources_to_patch,
            key=lambda x: x[1],
        )

        # Validate resources to be patched
        group_number = 1
        resource_number = 0
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for (
                group_name,
                group_weight,
                resource_list,
            ) in aws_resources_to_patch_sorted:
                print_subheading(f"\n{group_number}. {group_name}")
                resource_number += len(resource_list)
                group_number += 1
                for resource in resource_list:
                    print_info(
                        f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                    )
            if self.aws_args.aws_region:
                print_info(f"\nRegion: {self.aws_args.aws_region}")
            if self.aws_args.aws_profile:
                print_info(f"Profile: {self.aws_args.aws_profile}")
            print_info(f"\nTotal {resource_number} resources")
            confirm = confirm_yes_no("\nConfirm patch")
            if not confirm:
                print_info("Skipping patch")
                return False

        # track the total number of AwsResources to patch for validation
        num_resources_to_patch: int = 0
        num_resources_patched: int = 0

        for (group_name, group_weight, resource_list) in aws_resources_to_patch_sorted:
            print_subheading(f"\n-*- {group_name}")
            num_resources_to_patch += len(resource_list)
            for resource in resource_list:
                if resource:
                    print_info(
                        f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
                    )
                    # logger.debug(resource)
                    try:
                        _resource_patched = resource.update(
                            aws_client=self.aws_api_client
                        )
                        if _resource_patched:
                            num_resources_patched += 1
                    except Exception as e:
                        logger.error(
                            f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be patched."
                        )
                        logger.error("Error: {}".format(e))
                        logger.error(
                            "Skipping resource creation, please fix and try again..."
                        )

        print_info(
            f"\n# Resources patched: {num_resources_patched}/{num_resources_to_patch}"
        )
        if num_resources_to_patch == num_resources_patched:
            return True

        logger.error(
            f"Resources patched: {num_resources_patched} do not match resources required: {num_resources_to_patch}"
        )
        return False

    def patch_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Patching AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return

        # A list of tuples with 3 parts
        #   1. Resource group name
        #   2. Resource group weight
        #   3. List of Resources in group after filters
        aws_resources_to_patch: List[Tuple[str, int, List[AwsResource]]] = []
        for aws_rg_name, aws_rg in self.aws_resources.items():
            logger.debug(f"Processing: {aws_rg_name}")

            # skip disabled AwsResourceGroups
            if not aws_rg.enabled:
                continue

            # skip groups not matching app_filter if provided
            if app_filter is not None:
                if app_filter.lower() not in aws_rg_name.lower():
                    logger.debug(f"Skipping {aws_rg_name}")
                    continue

            aws_resources_in_group = get_aws_resources_from_group(
                aws_resource_group=aws_rg,
                name_filter=name_filter,
                type_filter=type_filter,
            )
            num_resources_in_group = len(aws_resources_in_group)
            logger.debug(f"Matched {num_resources_in_group} resources")
            if num_resources_in_group > 0:
                aws_resources_to_patch.append(
                    (aws_rg_name, aws_rg.weight, aws_resources_in_group)
                )

        if len(aws_resources_to_patch) == 0:
            print_info("No AwsResources to patch")
            return

        aws_resources_to_patch_sorted: List[
            Tuple[str, int, List[AwsResource]]
        ] = sorted(
            aws_resources_to_patch,
            key=lambda x: x[1],
        )

        group_number = 1
        num_resources_to_patch: int = 0
        print_heading("--**-- Aws resources:")
        for (group_name, group_weight, resource_list) in aws_resources_to_patch_sorted:
            print_subheading(f"\n{group_number}. {group_name}")
            num_resources_to_patch += len(resource_list)
            group_number += 1
            for resource in resource_list:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
        if self.aws_args.aws_region:
            print_info(f"\nRegion: {self.aws_args.aws_region}")
        if self.aws_args.aws_profile:
            print_info(f"Profile: {self.aws_args.aws_profile}")
        print_info(f"\nTotal {num_resources_to_patch} resources")

    ######################################################
    ## End
    ######################################################



from typing import Optional, Any, List, Dict

from botocore.exceptions import ClientError

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_info, print_error, print_warning
from phidata.utils.log import logger


class IamPolicy(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/iam.html#policy
    """

    resource_type = "IamPolicy"
    service_name = "iam"

    # PolicyName
    # The friendly name of the policy.
    name: str
    # The JSON policy document that you want to use as the content for the new policy.
    # You must provide policies in JSON format in IAM.
    # However, for CloudFormation templates formatted in YAML, you can provide the policy in JSON or YAML format.
    # CloudFormation always converts a YAML policy to JSON format before submitting it to IAM.
    policy_document: str
    # The path for the policy. This parameter is optional. If it is not included, it defaults to a slash (/).
    path: Optional[str] = None
    # A friendly description of the policy.
    description: Optional[str] = None
    # A list of tags that you want to attach to the new policy. Each tag consists of a key name and an associated value.
    tags: Optional[List[Dict[str, str]]] = None

    arn: Optional[str] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the IamPolicy

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # create a dict of args which are not null, otherwise aws type validation fails
            not_null_args: Dict[str, Any] = {}
            if self.path:
                not_null_args["Path"] = self.path
            if self.description:
                not_null_args["Description"] = self.description
            if self.tags:
                not_null_args["Tags"] = self.tags

            # Create Policy
            service_resource = self.get_service_resource(aws_client)
            policy = service_resource.create_policy(
                PolicyName=self.name,
                PolicyDocument=self.policy_document,
                **not_null_args,
            )
            # logger.debug(f"Policy: {policy}")

            # Validate Policy creation
            create_date = policy.create_date
            self.arn = policy.arn
            logger.debug(f"create_date: {create_date}")
            logger.debug(f"arn: {self.arn}")
            if create_date is not None:
                print_info(f"Policy created: {self.name}")
                self.active_resource = policy
                return True
            print_error("Policy could not be created")
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # Wait for Policy to be created
        if self.wait_for_creation:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be created.")
                if self.arn is not None:
                    waiter = self.get_service_client(aws_client).get_waiter(
                        "policy_exists"
                    )
                    waiter.wait(
                        PolicyArn=self.arn,
                        WaiterConfig={
                            "Delay": self.waiter_delay,
                            "MaxAttempts": self.waiter_max_attempts,
                        },
                    )
                else:
                    print_warning("Skipping waiter, No Policy ARN found")
            except Exception as e:
                print_error("Waiter failed.")
                print_error(e)
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the IamPolicy

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_resource = self.get_service_resource(aws_client)
            policy = None
            for _policy in service_resource.policies.all():
                if _policy.policy_name == self.name:
                    policy = _policy
                    break

            if policy is None:
                logger.debug("No Policy found")
                return None

            policy.load()
            create_date = policy.create_date
            self.arn = policy.arn
            logger.debug(f"create_date: {create_date}")
            logger.debug(f"arn: {self.arn}")
            if create_date is not None:
                logger.debug(f"Policy found: {policy.policy_name}")
                self.active_resource = policy
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
            pass
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the IamPolicy

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            policy = self._read(aws_client)
            # logger.debug(f"Policy: {policy}")
            # logger.debug(f"Policy type: {type(policy)}")
            self.active_resource = None

            if policy is None:
                logger.warning(f"No {self.get_resource_type()} to delete")
                return True

            # Before you can delete a managed policy,
            # you must first detach the policy from all users, groups, and roles
            # that it is attached to. In addition, you must delete all
            # the policy's versions.

            # detach all roles
            roles = policy.attached_roles.all()
            for role in roles:
                print_info(f"Detaching policy from role: {role}")
                policy.detach_role(RoleName=role.name)

            # detach all users
            users = policy.attached_users.all()
            for user in users:
                print_info(f"Detaching policy from user: {user}")
                policy.detach_user(UserName=user.name)

            # detach all groups
            groups = policy.attached_groups.all()
            for group in groups:
                print_info(f"Detaching policy from group: {group}")
                policy.detach_group(GroupName=group.name)

            # delete all versions
            default_version = policy.default_version
            versions = policy.versions.all()
            for version in versions:
                if version.version_id == default_version.version_id:
                    print_info(f"Skipping deleting default PolicyVersion: {version}")
                    continue
                print_info(f"Deleting PolicyVersion: {version}")
                version.delete()

            # delete policy
            policy.delete()
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
            )
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False



from typing import List, Optional, Any, Dict

from kubernetes.client.models.v1_container import V1Container
from kubernetes.client.models.v1_pod_spec import V1PodSpec
from kubernetes.client.models.v1_volume import V1Volume
from pydantic import Field

from phidata.infra.k8s.enums.restart_policy import RestartPolicy
from phidata.infra.k8s.resource.base import K8sObject
from phidata.infra.k8s.resource.core.v1.container import Container
from phidata.infra.k8s.resource.core.v1.toleration import Toleration
from phidata.infra.k8s.resource.core.v1.topology_spread_constraints import (
    TopologySpreadConstraint,
)
from phidata.infra.k8s.resource.core.v1.local_object_reference import (
    LocalObjectReference,
)
from phidata.infra.k8s.resource.core.v1.volume import Volume


class PodSpec(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#podspec-v1-core
    """

    resource_type: str = "PodSpec"

    # Optional duration in seconds the pod may be active on the node relative to StartTime before
    # the system will actively try to mark it failed and kill associated containers.
    # Value must be a positive integer.
    active_deadline_seconds: Optional[int] = Field(None, alias="activeDeadlineSeconds")
    # If specified, the pod's scheduling constraints
    # TODO: create affinity object
    affinity: Optional[Any] = None
    # AutomountServiceAccountToken indicates whether a service account token should be automatically mounted.
    automount_service_account_token: Optional[bool] = Field(
        None, alias="automountServiceAccountToken"
    )
    # List of containers belonging to the pod. Containers cannot currently be added or removed.
    # There must be at least one container in a Pod. Cannot be updated.
    containers: List[Container]
    # Specifies the DNS parameters of a pod.
    # Parameters specified here will be merged to the generated DNS configuration based on DNSPolicy.
    # TODO: create dns_config object
    dns_config: Optional[Any] = Field(None, alias="dnsConfig")
    dns_policy: Optional[str] = Field(None, alias="dnsPolicy")
    # ImagePullSecrets is an optional list of references to secrets in the same namespace to
    # use for pulling any of the images used by this PodSpec.
    # If specified, these secrets will be passed to individual puller implementations for them to use.
    # For example, in the case of docker, only DockerConfig type secrets are honored.
    # More info: https://kubernetes.io/docs/concepts/containers/images#specifying-imagepullsecrets-on-a-pod
    image_pull_secrets: Optional[List[LocalObjectReference]] = Field(
        None, alias="imagePullSecrets"
    )
    # List of initialization containers belonging to the pod.
    # Init containers are executed in order prior to containers being started.
    # If any init container fails, the pod is considered to have failed and is
    # handled according to its restartPolicy.
    # The name for an init container or normal container must be unique among all containers.
    # Init containers may not have Lifecycle actions, Readiness probes, Liveness probes, or Startup probes.
    # The resourceRequirements of an init container are taken into account during scheduling by finding
    # the highest request/limit for each resource type, and then using the max of that value or
    # the sum of the normal containers. Limits are applied to init containers in a similar fashion.
    # Init containers cannot currently be added or removed. Cannot be updated.
    # More info: https://kubernetes.io/docs/concepts/workloads/pods/init-containers/
    init_containers: Optional[List[Container]] = Field(None, alias="initContainers")
    # NodeName is a request to schedule this pod onto a specific node.
    # If it is non-empty, the scheduler simply schedules this pod onto that node,
    # assuming that it fits resource requirements.
    node_name: Optional[str] = Field(None, alias="nodeName")
    # NodeSelector is a selector which must be true for the pod to fit on a node.
    # Selector which must match a node's labels for the pod to be scheduled on that node.
    # More info: https://kubernetes.io/docs/concepts/configuration/assign-pod-node/
    node_selector: Optional[Dict[str, str]] = Field(None, alias="nodeSelector")
    # Restart policy for all containers within the pod.
    # One of Always, OnFailure, Never. Default to Always.
    # More info: https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle/#restart-policy
    restart_policy: Optional[RestartPolicy] = Field(None, alias="restartPolicy")
    # ServiceAccountName is the name of the ServiceAccount to use to run this pod.
    # More info: https://kubernetes.io/docs/tasks/configure-pod-container/configure-service-account/
    service_account_name: Optional[str] = Field(None, alias="serviceAccountName")
    termination_grace_period_seconds: Optional[int] = Field(
        None, alias="terminationGracePeriodSeconds"
    )
    # If specified, the pod's tolerations.
    tolerations: Optional[List[Toleration]] = None
    # TopologySpreadConstraints describes how a group of pods ought to spread across topology domains.
    # Scheduler will schedule pods in a way which abides by the constraints.
    # All topologySpreadConstraints are ANDed.
    topology_spread_constraints: Optional[List[TopologySpreadConstraint]] = Field(
        None, alias="topologySpreadConstraints"
    )
    # List of volumes that can be mounted by containers belonging to the pod.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes
    volumes: Optional[List[Volume]] = None

    def get_k8s_object(self) -> V1PodSpec:
        # Set and return a V1PodSpec object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_pod_spec.py

        _containers: Optional[List[V1Container]] = None
        if self.containers:
            _containers = []
            for _container in self.containers:
                _containers.append(_container.get_k8s_object())

        _init_containers: Optional[List[V1Container]] = None
        if self.init_containers:
            _init_containers = []
            for _init_container in self.init_containers:
                _init_containers.append(_init_container.get_k8s_object())

        _image_pull_secrets = None
        if self.image_pull_secrets:
            _image_pull_secrets = []
            for ips in self.image_pull_secrets:
                _image_pull_secrets.append(ips.get_k8s_object())

        _volumes: Optional[List[V1Volume]] = None
        if self.volumes:
            _volumes = []
            for _volume in self.volumes:
                _volumes.append(_volume.get_k8s_object())

        _v1_pod_spec = V1PodSpec(
            active_deadline_seconds=self.active_deadline_seconds,
            affinity=self.affinity,
            automount_service_account_token=self.automount_service_account_token,
            containers=_containers,
            dns_config=self.dns_config,
            dns_policy=self.dns_policy,
            image_pull_secrets=_image_pull_secrets,
            init_containers=_init_containers,
            node_name=self.node_name,
            node_selector=self.node_selector,
            restart_policy=self.restart_policy,
            service_account_name=self.service_account_name,
            termination_grace_period_seconds=self.termination_grace_period_seconds,
            volumes=_volumes,
        )
        return _v1_pod_spec



from pathlib import Path
from typing import Optional, Union

from phidata.app.k8s.app import K8sApp, K8sAppArgs
from phidata.exceptions.app import AppInvalidException
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.log import logger


class K8sDirArgs(K8sAppArgs):
    dir: Union[str, Path]


class K8sDir(K8sApp):
    def __init__(
        self,
        dir: Union[str, Path],
        name: Optional[str] = None,
        version: str = "1",
        enabled: bool = True,
        cache_manifests: bool = True,
    ):
        super().__init__()

        try:
            self.args: K8sDirArgs = K8sDirArgs(
                dir=dir,
                name=name,
                version=version,
                enabled=enabled,
                cache_manifests=cache_manifests,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def get_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:

        # Validate URL
        manifest_dir = self.args.dir
        if manifest_dir is None:
            raise AppInvalidException("Invalid dir")

        if isinstance(manifest_dir, str):
            workspace_config_dir = self.workspace_config_dir
            if workspace_config_dir:
                self.args.manifest_dir = (
                    Path(workspace_config_dir).joinpath(manifest_dir).resolve()
                )
        elif isinstance(manifest_dir, Path):
            self.args.manifest_dir = manifest_dir

        if self.args.name is None:
            self.args.name = self.args.manifest_dir.stem

        return super().get_k8s_rg(k8s_build_context=k8s_build_context)
